# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objects as go
from crystal_toolkit.core.mpcomponent import MPComponent
from crystal_toolkit.helpers.layouts import Columns, Column
from dash_core_components import Loading
from plotly.subplots import make_subplots
from vise.analyzer.dos_data import DosPlotData
from vise.analyzer.plot_band import BandPlotInfo


class BandDosComponent(MPComponent):
    def __init__(self,
                 dos_plot_data: DosPlotData,
                 band_plot_info: BandPlotInfo,
                 band_plot_info_2: BandPlotInfo = None,
                 id: str = None,
                 *args, **kwargs):
        super().__init__(id=id, *args, **kwargs)
        self.dos_plot_data = dos_plot_data
        self.band_plot_info = band_plot_info
        self.band_plot_info_2 = band_plot_info_2
#        self.create_store("dos_data", initial_data=dos_data)

    @property
    def _sub_layouts(self):

        # pdos_range = self.get_slider_input(
        #     kwarg_label="pdos_range",
        #     state={},
        #     label="PDOS max range",
        #     min=0.1,
        #     max=
        # )

        if self.dos_plot_data:
            num_dos = len(self.dos_plot_data.doses)
            names = ["band"] + self.dos_plot_data.names
            column_width = [0.25] + [0.75 / num_dos] * num_dos
        else:
            num_dos = 1
            names = ["band"]
            column_width = [0.4, 0.6]

        fig = make_subplots(rows=1,
                            cols=1 + num_dos,
                            column_widths=column_width,
                            shared_yaxes=True,
                            horizontal_spacing=0.01,
                            subplot_titles=names,
                            font_size=25,
                            x_title="DOS (1/eV/unit cell)",
                            y_title="Energy (eV)")

        if self.dos_plot_data:
            fig["layout"]["yaxis"]["range"] = self.dos_plot_data.energy_range
        else:
            fig["layout"]["yaxis"]["range"] = [-5, 10]

        if self.band_plot_info:
            self.add_band(fig)
        else:
            fig.add_trace(
                go.Scatter(x=[], y=[], name="band"), row=1, col=1)

        fig.update_layout(width=2000, height=700)

        if self.dos_plot_data:
            self.add_dos(fig)

        # Main plot
        graph = Loading([dcc.Graph(figure=fig)])

        return {"graph": graph}

    def add_band(self, fig):
        last_path = self.band_plot_info.x_ticks.distances[-1]
        fig["layout"]["xaxis"]["range"] = [0.0, last_path]

        band_edge = self.band_plot_info.band_info_set[0].band_edge
        h_lines_2 = None

        if band_edge:
            base_energy = band_edge.vbm
            cbm = band_edge.cbm - base_energy
            h_lines = [0.0, cbm]
            # Add vbm and cbm.
            fig.add_trace(
                go.Scatter(x=band_edge.vbm_distances,
                           y=[0.0] * len(band_edge.vbm_distances),
                           line_color="blue",
                           fillcolor="blue",
                           mode="markers",
                           marker_size=20,
                           opacity=0.7,
                           showlegend=False),
                row=1, col=1),
            fig.add_trace(
                go.Scatter(x=band_edge.cbm_distances,
                           y=[cbm] * len(band_edge.cbm_distances),
                           line_color="red",
                           fillcolor="red",
                           mode="markers",
                           marker_size=20,
                           opacity=0.7,
                           showlegend=False),
                row=1, col=1),

            if self.band_plot_info_2:
                band_edge_2 = self.band_plot_info_2.band_info_set[0].band_edge
                vbm = band_edge_2.vbm - base_energy
                cbm = band_edge_2.cbm - base_energy
                h_lines_2 = [vbm, cbm]
                # Add vbm and cbm.
                fig.add_trace(
                    go.Scatter(x=band_edge_2.vbm_distances,
                               y=[vbm] * len(band_edge_2.vbm_distances),
                               line_color="blue",
                               fillcolor="blue",
                               mode="markers",
                               marker_size=20,
                               opacity=0.5,
                               showlegend=False),
                    row=1, col=1),
                fig.add_trace(
                    go.Scatter(x=band_edge_2.cbm_distances,
                               y=[cbm] * len(band_edge_2.cbm_distances),
                               line_color="red",
                               fillcolor="red",
                               mode="markers",
                               marker_size=20,
                               opacity=0.5,
                               showlegend=False),
                    row=1, col=1),

        else:
            base_energy = self.band_plot_info.band_info_set[0].fermi_level
            h_lines = [0.0]

        # Add vbm and cbm lines.
        for energy in h_lines:
            fig.add_shape(dict(type="line",
                               x0=0, x1=last_path, y0=energy, y1=energy,
                               line=dict(color="royalblue", width=3, dash="dash")),
                          row=1, col=1)

        if h_lines_2:
            for energy in h_lines:
                fig.add_shape(dict(type="line",
                                   x0=0, x1=last_path, y0=energy, y1=energy,
                                   line=dict(color="royalblue", width=3, dash="dot")),
                              row=1, col=1)


        # Add x labels
        str_replace = {
            "$\mid$": "|",
            "_1": "₁",
            "_2": "₂",
            "_3": "₃",
            "_4": "₄",
            r"${\rm ": "",
            "}": "",
            "$": "",
        }
        labels = self.band_plot_info.x_ticks.labels
        distances = self.band_plot_info.x_ticks.distances
        new_labels = []
        for label in labels:
            for key in str_replace.keys():
                if key in label:
                    label = label.replace(key, str_replace[key])
            new_labels.append(label)

        fig.update_xaxes(tickvals=self.band_plot_info.x_ticks.distances,
                         ticktext=new_labels,
                         tickfont_size=16,
                         row=1, col=1)

        # Add k-path border lines.
        for label, distance in zip(new_labels, distances):
            if "|" in label:
                fig.add_shape(dict(type="line",
                                   x0=distance, x1=distance, y0=-20, y1=20,
                                   line=dict(color="black",
                                             width=2)), row=1, col=1)

        for d, e in zip(self.band_plot_info.distances_by_branch,
                        self.band_plot_info.band_info_set[0].band_energies):
            for i in e[0]:
                fig.add_trace(
                    go.Scatter(x=d,
                               y=[j - base_energy for j in i],
                               hoverinfo="skip",
                               line_color="black",
                               showlegend=False,
                               name="band",
                               mode="lines",
                               line={"width": 2.0}), row=1, col=1),

        if self.band_plot_info_2:
            for d, e in zip(self.band_plot_info.distances_by_branch,
                            self.band_plot_info_2.band_info_set[0].band_energies):
                for i in e[0]:
                    fig.add_trace(
                        go.Scatter(x=d,
                                   y=[j - base_energy for j in i],
                                   hoverinfo="skip",
                                   line_color="green",
                                   showlegend=False,
                                   name="band",
                                   mode="lines",
                                   opacity=0.5,
                                   line={"width": 3.0}), row=1, col=1),

    def add_dos(self, fig):
        total_dos = self.dos_plot_data.doses[0][0].dos[0]
        energies = self.dos_plot_data.relative_energies
        cbm = min([e for d, e in zip(total_dos, energies)
                   if e > 0 and d > 1e-5])

        colors = ["#1f77b4",  # muted blue
                  "#ff7f0e",  # safety orange
                  "#2ca02c",  # cooked asparagus green
                  "#9467bd",  # muted purple
                  ]
        for i, (doses, y_lim) in enumerate(zip(self.dos_plot_data.doses,
                                               self.dos_plot_data.dos_ranges), 2):
            for dos, color in zip(doses, colors):
                fig.add_trace(
                    go.Scatter(y=self.dos_plot_data.relative_energies,
                               x=dos.dos[0],
                               name=dos.name,
                               line_color=color,
                               line={"width": 3},
                               legendgroup="total" if i == 2 else "pdos",
                               showlegend=True if i == 3 else False),
                    row=1, col=i
                )
            for energy in [0.0, cbm]:
                fig.add_shape(dict(type="line",
                                   x0=y_lim[0], x1=y_lim[1],
                                   y0=energy, y1=energy,
                                   line=dict(color="royalblue",
                                             width=3,
                                             dash="dash")), row=1, col=i)
            fig.update_xaxes(range=y_lim, row=1, col=i)
        fig.update_layout(legend=dict(yanchor="top", y=0.99,
                                      xanchor="right", x=0.99))

    def layout(self):
        return html.Div([Columns(Column(self._sub_layouts["graph"]))])

    def generate_callbacks(self, app, cache):
        pass
    #     @app.callback(
    #         Output(self.id("dos-plot"), "figure"),
    #         [
    #             Input(self.get_kwarg_id("grouping"), "value"),
    #         ],
    #     )
    #     def update_graph(grouping):

            # plot_data: DosPlotData = self.dos_plot_data.dos_plot_data({"O": [10, 11]})

            # fig = make_subplots(rows=1, cols=2, shared_yaxes=True,
            #                     horizontal_spacing=0.01)

            # fig.add_trace(
            #     go.Scatter(y=plot_data.relative_energies,
            #                x=plot_data.doses[0][0].dos[0],
            #                name="total", line={"width": 4}),
            #     row=1, col=1
            # )

            # fig.add_trace(
            #     go.Scatter(y=plot_data.relative_energies,
            #                x=plot_data.doses[1][0].dos[0],
            #                name="s", line={"width": 4}),
            #     row=1, col=2
            # )

            # fig.add_trace(
            #     go.Scatter(y=plot_data.relative_energies,
            #                x=plot_data.doses[1][1].dos[0],
            #                name="p", line={"width": 4}),
            #     row=1, col=2
            # )

            # fig.add_trace(
            #     go.Scatter(y=plot_data.relative_energies,
            #                x=plot_data.doses[1][2].dos[0],
            #                name="d", line={"width": 4}),
            #     row=1, col=2
            # )

            # return fig




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from math import pi, sqrt

import numpy as np
import pytest
from monty.serialization import loadfn
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.dielectric_function import DieleFuncData, \
    eV_to_inv_cm
from vise.analyzer.vasp.make_diele_func import make_diele_func
from vise.tests.conftest import assert_msonable

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


@pytest.fixture
def diele_func_data():
    array_real = [1, 2, 3, 0, 0, 0]
    array_imag = [4, 5, 6, 0, 0, 0]
    return DieleFuncData(energies=list(np.linspace(0.0, 10.0, num=11)),
                         diele_func_real=[array_real]*11,
                         diele_func_imag=[array_imag]*11,
                         band_gap=1.0)


def test_msonable(diele_func_data):
    assert_msonable(diele_func_data)


def test_json_file_mixin(diele_func_data, tmpdir):
    tmpdir.chdir()
    diele_func_data.to_json_file()
    actual = loadfn("diele_func_data.json")
    assert actual.diele_func_real == diele_func_data.diele_func_real


def test_ave_absorption_coeff(diele_func_data):
    actual = diele_func_data.ave_absorption_coeff
    expected = [2 * sqrt(2) * pi * sqrt(sqrt(2 ** 2 + 5 ** 2) - 2)
                * i * eV_to_inv_cm for i in range(0, 11)]
    assert actual == expected


@pytest.fixture
def actual_diele_func_data(test_data_files):
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun.xml")
    o = Outcar(test_data_files / "MgSe_absorption_OUTCAR")
    return make_diele_func(v, o)


def test_target_coeff_e_from_band_gap(actual_diele_func_data):
    actual = actual_diele_func_data.target_coeff_e_from_band_gap
    np.testing.assert_almost_equal(actual, 0.4014000000000002)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from monty.serialization import loadfn
from vise.analyzer.plot_band_dos import BandDosPlotlyPlotter
from vise.util.dash_helper import show_png

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test(test_data_files):
    band_plot_data = loadfn(str(test_data_files / "KAlSi3O8_band_plot_info.json"))
    dos_plot_data = loadfn(str(test_data_files / "KAlSi3O8_dos_plot_data.json"))
    band_dos_component = BandDosPlotlyPlotter(dos_plot_data, band_plot_data)
#    band_dos_component.fig.show()
    show_png(band_dos_component.fig)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.plot_absorption_coeff import AbsorptionCoeffPlotlyPlotter, \
    AbsorptionCoeffMplPlotter

from vise.analyzer.vasp.make_diele_func import make_diele_func
from vise.util.dash_helper import show_png


try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


@pytest.fixture
def actual_diele_func_data(test_data_files):
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun.xml")
    o = Outcar(test_data_files / "MgSe_absorption_OUTCAR")
    return make_diele_func(v, o)


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test_absorption_coeff_plotly_plotter(actual_diele_func_data):
    plotter = AbsorptionCoeffPlotlyPlotter(actual_diele_func_data,
                                           materials=["GaAs"])
    fig = plotter.create_figure()
    show_png(fig)


def test_absorption_coeff_mpl_plotter(actual_diele_func_data):
    plotter = AbsorptionCoeffMplPlotter(actual_diele_func_data,
                                        materials=["GaAs", "Si"])
    #                                        align_gap=False)
    plotter.construct_plot()
    plotter.plt.show()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

import pytest
import dash
from crystal_toolkit.settings import SETTINGS
from dash import Dash
from pymatgen import loadfn
from vise.analyzer.dash_components.band_dos_dash import BandDosComponent
from vise.analyzer.dos_data import DosData
import dash_html_components as html

import crystal_toolkit.components as ctc

#
#def test_plotly_actual_plot(test_files):
#     dos_data = loadfn(str(test_files / "KAlSi3O8_dos_data.json"))
#     pploter = PlotlyDosPlotter(dos_data)
#     fig = pploter.show({"O": [10, 11]})
#     fig.show()


app = Dash(suppress_callback_exceptions=True,
           assets_folder=SETTINGS.ASSETS_PATH)
ctc.register_app(app)

test_files = Path(__file__).parent / ".." / ".." / "tests" / "test_data_files"
band_plot_data = loadfn(str(test_files / "KAlSi3O8_band_plot_info.json"))
dos_plot_data = loadfn(str(test_files / "KAlSi3O8_dos_plot_data.json"))
band_dos_component = BandDosComponent(dos_plot_data, band_plot_data)
#dos_component = BandDosComponent(None, band_plot_info)
#dos_component = BandDosComponent(dos_plot_data, band_plot_info, band_plot_info_2)

# example layout to demonstrate capabilities of component
my_layout = html.Div(
    [
        html.H1("BandDosComponent Example"),
        band_dos_component.layout(),
    ]
)
# {'K_i1': [0, 1],
#  'Al_i1': [2, 3],
#  'Si_i1': [4, 5],
#  'Si_i2': [6, 7],
#  'Si_i3': [8, 9],
#  'O_i1': [10, 11],
#  'O_i2': [12, 13],
#  'O_i3': [14, 15],
#  'O_i4': [16, 17],
#  'O_i5': [18, 19],
#  'O_i6': [20, 21],
#  'O_i7': [22, 23],
#  'O_i8': [24, 25]}
    # wrap your app.layout with crystal_toolkit_layout()
    # to ensure all necessary components are loaded into layout
# app.layout = ctc.crystal_toolkit_layout(my_layout)
ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)

    # allow app to be run using "python structure.py"
    # in production, deploy behind gunicorn or similar
    # see Dash documentation for more information
if __name__ == "__main__":
    app.run_server(debug=True, port=8051)


"""
TODO
- Enable to select inequivalent sites shown
- Enable to change the max range of dos

DONE
- Show vbm and cbm
- Show both pbesol and dd-hybrid.
- Change size of dos (smaller width)
- Set max range of default DOS.
- Add lines 
- Add band figure
- Add x and y labels
- title
- show s, p, d orbitals
- show multiple figures for K, Al, Si, and O
"""


# coding: utf-8
#  Copyright (c) 2020 Kumagai group.
import re
from collections import OrderedDict
from itertools import combinations_with_replacement, chain
from typing import Optional, Tuple

import dash_table as dt
import numpy as np
from crystal_toolkit.core.legend import Legend
from crystal_toolkit.core.mpcomponent import MPComponent
from crystal_toolkit.core.scene import Scene
from crystal_toolkit.helpers.layouts import *
from crystal_toolkit.settings import SETTINGS
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash_mp_components import Simple3DScene
from pymatgen.analysis.graphs import StructureGraph, MoleculeGraph
from pymatgen.analysis.local_env import NearNeighbors
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import DummySpecie
from pymatgen.core.structure import Structure, Molecule
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

DEFAULTS = {
    "bonding_strategy": "CrystalNN",
    "radius_strategy": "specified_or_average_ionic",
    "draw_image_atoms": True,
    "bonded_sites_outside_unit_cell": False,
    "hide_incomplete_bonds": True,
    "show_compass": True,
    "unit_cell_choice": "input",
}


class StructureComponent(MPComponent):
    """
    A component to display pymatgen Structure objects.
    """

    available_bonding_strategies = \
        {subcls.__name__: subcls for subcls in NearNeighbors.__subclasses__()}

    default_scene_settings = {
        "extractAxis": True,
        # For visual diff testing, we change the renderer
        # to SVG since this WebGL support is more difficult
        # in headless browsers / CI.
        "renderer": "svg" if SETTINGS.TEST_MODE else "webgl",
    }

    default_title = "Structure"

    def __init__(
            self,
            structure: Optional[Structure] = None,
            id: str = None,
            scene_additions: Optional[Scene] = None,
            bonding_strategy: str = DEFAULTS["bonding_strategy"],
            bonding_strategy_kwargs: Optional[dict] = None,
            color_scale: Optional[str] = None,
            radius_strategy: str = DEFAULTS["radius_strategy"],
            unit_cell_choice: str = DEFAULTS["unit_cell_choice"],
            draw_image_atoms: bool = DEFAULTS["draw_image_atoms"],
            bonded_sites_outside_unit_cell: bool = DEFAULTS[
                "bonded_sites_outside_unit_cell"
            ],
            hide_incomplete_bonds: bool = DEFAULTS["hide_incomplete_bonds"],
            show_compass: bool = DEFAULTS["show_compass"],
            scene_settings: Optional[Dict] = None,
            **kwargs,
    ):

        super().__init__(id=id, default_data=structure, **kwargs)

        self.initial_scene_settings = self.default_scene_settings.copy()
        if scene_settings:
            self.initial_scene_settings.update(scene_settings)

        self.create_store("scene_settings",
                          initial_data=self.initial_scene_settings)

        # unit cell choice and bonding algorithms need to come from a settings
        # object (in a dcc.Store) guaranteed to be present in layout, rather
        # than from the controls themselves -- since these are optional and
        # may not be present in the layout
        self.create_store(
            "graph_generation_options",
            initial_data={"bonding_strategy": bonding_strategy,
                          "bonding_strategy_kwargs": bonding_strategy_kwargs,
                          "unit_cell_choice": unit_cell_choice,
            },
        )

        self.create_store(
            "display_options",
            initial_data={
                "color_scale"                   : color_scale,
                "radius_strategy"               : radius_strategy,
                "draw_image_atoms"              : draw_image_atoms,
                "bonded_sites_outside_unit_cell": bonded_sites_outside_unit_cell,
                "hide_incomplete_bonds"         : hide_incomplete_bonds,
                "show_compass"                  : show_compass,
            },
        )

        if scene_additions:
            initial_scene_additions = Scene(name="scene_additions",
                                            contents=scene_additions).to_json()
        else:
            initial_scene_additions = None
        self.create_store("scene_additions",
                          initial_data=initial_scene_additions)

        if structure:
            # graph is cached explicitly, this isn't necessary but is an
            # optimization so that graph is only re-generated if bonding
            # algorithm changes
            graph = self._preprocess_input_to_graph(
                structure,
                bonding_strategy=bonding_strategy,
                bonding_strategy_kwargs=bonding_strategy_kwargs,
            )
            scene, legend = self.get_scene_and_legend(
                graph,
                scene_additions=self.initial_data["scene_additions"],
                **self.initial_data["display_options"],
            )
            if hasattr(structure, "lattice"):
                self._lattice = structure.lattice
        else:
            # component could be initialized without a structure, in which case
            # an empty scene should be displayed
            graph = None
            scene, legend = self.get_scene_and_legend(
                None,
                scene_additions=self.initial_data["scene_additions"],
                **self.initial_data["display_options"])

        self.create_store("legend_data", initial_data=legend)
        self.create_store("graph", initial_data=graph)

        # this is used by a Simple3DScene component, not a dcc.Store
        self._initial_data["scene"] = scene

        # hide axes inset for molecules
        if isinstance(structure, Molecule) or isinstance(
                structure, MoleculeGraph
        ):
            self.scene_kwargs = {"axisView": "HIDDEN"}
        else:
            self.scene_kwargs = {}

    def _make_legend(self, legend):

        if not legend:
            return html.Div(id=self.id("legend"))

        def get_font_color(hex_code):
            # ensures contrasting font color for background color
            c = tuple(int(hex_code[1:][i: i + 2], 16) for i in (0, 2, 4))
            if 1 - (c[0] * 0.299 + c[1] * 0.587 + c[2] * 0.114) / 255 < 0.5:
                font_color = "#000000"
            else:
                font_color = "#ffffff"
            return font_color

        try:
            formula = Composition.from_dict(
                legend["composition"]).reduced_formula
        except:
            # TODO: fix legend for Dummy Specie compositions
            formula = "Unknown"

        legend_colors = OrderedDict(
            sorted(list(legend["colors"].items()),
                   key=lambda x: formula.find(x[1]))
        )

        legend_elements = [
            Button(
                html.Span(
                    name, className="icon",
                    style={"color": get_font_color(color)}
                ),
                kind="static",
                style={"backgroundColor": color},
            )
            for color, name in legend_colors.items()
        ]

        return Field(
            [Control(el, style={"marginRight": "0.2rem"}) for el in
             legend_elements],
            id=self.id("legend"),
            grouped=True,
        )

    def _make_title(self, legend):

        if not legend or (not legend.get("composition", None)):
            return H1(self.default_title, id=self.id("title"))

        composition = legend["composition"]
        if isinstance(composition, dict):

            try:
                composition = Composition.from_dict(composition)

                # strip DummySpecie if present (TODO: should be method in pymatgen)
                composition = Composition(
                    {
                        el: amt
                        for el, amt in composition.items()
                        if not isinstance(el, DummySpecie)
                    }
                )
                composition = composition.get_reduced_composition_and_factor()[
                    0]
                formula = composition.reduced_formula
                formula_parts = re.findall(r"[^\d_]+|\d+", formula)
                formula_components = [
                    html.Sub(part.strip())
                    if part.isnumeric()
                    else html.Span(part.strip())
                    for part in formula_parts
                ]
            except:
                formula_components = list(map(str, composition.keys()))

        return H1(
            formula_components, id=self.id("title"),
            style={"display": "inline-block"}
        )

    @staticmethod
    def _make_bonding_algorithm_custom_cuffoff_data(graph):
        if not graph:
            return [{"A": None, "B": None, "A—B": None}]
        struct_or_mol = StructureComponent._get_structure(graph)
        # can't use type_of_specie because it doesn't work with disordered structures
        species = set(
            map(
                str,
                chain.from_iterable(
                    [list(c.keys()) for c in struct_or_mol.species_and_occu]
                ),
            )
        )
        rows = [
            {"A": combination[0], "B": combination[1], "A—B": 0}
            for combination in combinations_with_replacement(species, 2)
        ]
        return rows

    @property
    def _sub_layouts(self):

        struct_layout = html.Div(
            Simple3DScene(
                id=self.id("scene"),
                data=self.initial_data["scene"],
                settings=self.initial_scene_settings,
                sceneSize="100%",
                **self.scene_kwargs,
            ),
            style={
                "width"   : "100%",
                "height"  : "100%",
                "overflow": "hidden",
                "margin"  : "0 auto",
            },
        )

        title_layout = html.Div(
            self._make_title(self._initial_data["legend_data"]),
            id=self.id("title_container"),
        )

        legend_layout = html.Div(
            self._make_legend(self._initial_data["legend_data"]),
            id=self.id("legend_container"),
        )

        return {"struct": struct_layout,
                "title": title_layout,
                "legend": legend_layout}

    def layout(self, size: str = "500px") -> html.Div:
        """
        :param size: a CSS string specifying width/height of Div
        :return: A html.Div containing the 3D structure or molecule
        """
        return html.Div(
            self._sub_layouts["struct"], style={"width": size, "height": size}
        )

    @staticmethod
    def _preprocess_input_to_graph(
            input: Union[Structure, StructureGraph],
            bonding_strategy: str = DEFAULTS["bonding_strategy"],
            bonding_strategy_kwargs: Optional[Dict] = None,
    ) -> Union[StructureGraph, MoleculeGraph]:

        # ensure fractional co-ordinates are normalized to be in [0,1)
        # (this is actually not guaranteed by Structure)
        try:
            input = input.as_dict(verbosity=0)
        except TypeError:
            # TODO: remove this, necessary for Slab(?), some structure subclasses don't have verbosity
            input = input.as_dict()
        for site in input["sites"]:
            site["abc"] = np.mod(site["abc"], 1)
        input = Structure.from_dict(input)

        if not input.is_ordered:
            # calculating bonds in disordered structures is currently very flaky
            bonding_strategy = "CutOffDictNN"

        if (bonding_strategy
                not in StructureComponent.available_bonding_strategies.keys()):
            raise ValueError(
                "Bonding strategy not supported. Please supply a name "
                "of a NearNeighbor subclass, choose from: {}".format(
                    ", ".join(
                        StructureComponent.available_bonding_strategies.keys()
                    )
                )
            )
        else:
            bonding_strategy_kwargs = bonding_strategy_kwargs or {}
            if bonding_strategy == "CutOffDictNN":
                if "cut_off_dict" in bonding_strategy_kwargs:
                    # TODO: remove this hack by making args properly JSON serializable
                    bonding_strategy_kwargs["cut_off_dict"] = {
                        (x[0], x[1]): x[2]
                        for x in bonding_strategy_kwargs["cut_off_dict"]
                    }
            bonding_strategy = StructureComponent.available_bonding_strategies[bonding_strategy](**bonding_strategy_kwargs)
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    graph = StructureGraph.with_local_env_strategy(input, bonding_strategy)
            except:
                # for some reason computing bonds failed, so let's not have any bonds(!)
                graph = StructureGraph.with_empty_graph(input)

        return graph

    def generate_callbacks(self, app, cache):
        pass

    @staticmethod
    def _get_structure(graph: Union[StructureGraph, Structure]) -> Union[Structure, Molecule]:
        if isinstance(graph, StructureGraph):
            return graph.structure
        elif isinstance(graph, Structure):
            return graph
        else:
            raise ValueError

    @staticmethod
    def get_scene_and_legend(
            graph: Optional[Union[StructureGraph, MoleculeGraph]],
            color_scale=None,
            radius_strategy=DEFAULTS["radius_strategy"],
            draw_image_atoms=DEFAULTS["draw_image_atoms"],
            bonded_sites_outside_unit_cell=DEFAULTS[
                "bonded_sites_outside_unit_cell"],
            hide_incomplete_bonds=DEFAULTS["hide_incomplete_bonds"],
            explicitly_calculate_polyhedra_hull=False,
            scene_additions=None,
            show_compass=DEFAULTS["show_compass"],
    ) -> Tuple[Scene, Dict[str, str]]:

        scene = Scene(name="AceStructureMoleculeComponentScene")

        if graph is None:
            return scene, {}

        structure = StructureComponent._get_structure(graph)

        # TODO: add radius_scale
        legend = Legend(
            structure,
            color_scheme="VESTA",
            radius_scheme=radius_strategy,
            cmap_range=color_scale,
        )

        scene = graph.get_scene(
                draw_image_atoms=draw_image_atoms,
                bonded_sites_outside_unit_cell=bonded_sites_outside_unit_cell,
                hide_incomplete_edges=hide_incomplete_bonds,
                explicitly_calculate_polyhedra_hull=explicitly_calculate_polyhedra_hull,
                legend=legend,
            )

        scene.name = "StructureComponentScene"

        if hasattr(structure, "lattice"):
            axes = structure.lattice._axes_from_lattice()
            axes.visible = show_compass
            scene.contents.append(axes)

        scene = scene.to_json()
        if scene_additions:
            # TODO: need a Scene.from_json() to make this work
            # raise NotImplementedError
            scene["contents"].append(scene_additions)

        return scene, legend.get_legend()

    def title_layout(self):
        """
        :return: A layout including the composition of the structure/molecule as a title.
        """
        return self._sub_layouts["title"]

    def legend_layout(self):
        """
        :return: A layout including a legend for the structure/molecule.
        """
        return self._sub_layouts["legend"]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import copy
from dataclasses import dataclass
from functools import reduce
from typing import Dict, Optional, List

import numpy as np
from monty.json import MSONable
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class PDos(MSONable):
    s:   np.ndarray  # [by spin][by energy]
    px:  np.ndarray
    py:  np.ndarray
    pz:  np.ndarray
    dxy: np.ndarray
    dyz: np.ndarray
    dxz: np.ndarray
    dx2: np.ndarray
    dz2: np.ndarray
    f_3: Optional[np.ndarray] = None
    f_2: Optional[np.ndarray] = None
    f_1: Optional[np.ndarray] = None
    f0:  Optional[np.ndarray] = None
    f1:  Optional[np.ndarray] = None
    f2:  Optional[np.ndarray] = None
    f3:  Optional[np.ndarray] = None

    @property
    def p(self):
        return self.px + self.py + self.pz

    @property
    def d(self):
        return self.dxy + self.dyz + self.dxz + self.dx2 + self.dz2

    @property
    def f(self):
        try:
            return (self.f_3 + self.f_2 + self.f_1 +
                    self.f0 + self.f1 + self.f2 + self.f3)
        except TypeError:
            return None

    def __add__(self, other: "PDos"):
        args = {}
        for k, v in self.__dict__.items():
            if v is None:
                args[k] = None
            else:
                args[k] = v + getattr(other, k)
        return PDos(**args)


@dataclass
class DosData(MSONable):
    energies: List[float]
    total: np.ndarray
    pdos: List[PDos]
    vertical_lines: List[float]
    base_energy: Optional[float] = 0.0

    @property
    def spin(self):
        return False if len(self.total) == 1 else True

    def dos_plot_data(self,
                      grouped_atom_indices: Dict[str, List[int]],
                      xlim: Optional[List[float]] = None,
                      ylim_set: Optional[List[List[float]]] = None,
                      ) -> "DosPlotData":

        if ylim_set is not None:
            assert len(grouped_atom_indices) + 1 == len(ylim_set)  # total+pdos

        # Total dos does not have spin decomposition.
        doses = [[DosBySpinEnergy("", self.total.tolist())]]
        names = ["total"]

        for name, atom_indices_by_group in grouped_atom_indices.items():
            pdos_list = [self.pdos[idx] for idx in atom_indices_by_group]
            pdos = reduce(lambda x, y: x + y, pdos_list)
            pdos_by_ax = [DosBySpinEnergy("s", pdos.s.tolist()),
                          DosBySpinEnergy("p", pdos.p.tolist()),
                          DosBySpinEnergy("d", pdos.d.tolist())]
            if pdos.f is not None:
                pdos_by_ax.append(DosBySpinEnergy("f", pdos.f.tolist()))

            doses.append(pdos_by_ax)
            names.append(name)

        xlim = xlim or [-5, 10]
        abs_xlim = [x + self.base_energy for x in xlim]

        if ylim_set is None:
            if self.spin:
                ylim_set = [[-y, y] for y in self.max_y_ranges(doses, abs_xlim)]
            else:
                ylim_set = [[0, y] for y in self.max_y_ranges(doses, abs_xlim)]

        energies = [e - self.base_energy for e in self.energies]
        shifted_vertical_lines = [e - self.base_energy
                                  for e in self.vertical_lines]

        return DosPlotData(energies, doses, names, xlim, ylim_set,
                           shifted_vertical_lines)

    def max_y_ranges(self, doses, xlim, multi=1.1, round_digit=2):
        mask = np.ma.masked_outside(self.energies, xlim[0], xlim[1]).mask
        max_dos_by_ax = []
        for dos_by_ax in doses:
            max_dos_by_ax.append(np.max([dos_by_spin.max_dos(mask)
                                         for dos_by_spin in dos_by_ax]))

        total_dos_max = max_dos_by_ax[0]
        pdos_max = max(max_dos_by_ax[1:])
        plot_maxes = [total_dos_max] + [pdos_max] * (len(doses) - 1)

        return [round(i * multi, round_digit) for i in plot_maxes]


@dataclass
class DosBySpinEnergy(MSONable):
    name: str
    dos: List[List[float]]  # [by spin][by energy]

    def max_dos(self, mask: List[bool] = None):
        return max([np.max(np.ma.masked_array(d, mask).compressed())
                    for d in self.dos])


@dataclass
class DosPlotData(MSONable, ToJsonFileMixIn):
    relative_energies: List[float]
    doses: List[List[DosBySpinEnergy]]  # [by ax][by orbital]
    names: List[str]
    energy_range: List[float]
    dos_ranges: List[List[float]]
    energy_lines: List[float]

    @classmethod
    def from_dict(cls, d):
        for k in copy(d):
            if k[0] == "@":
                d.pop(k)

        if "xlim" in d:
            d["energy_range"] = d.pop("xlim")
        if "ylim_set" in d:
            d["dos_ranges"] = d.pop("ylim_set")
        if "vertical_lines" in d:
            d["energy_lines"] = d.pop("vertical_lines")
        for i, x in enumerate(d["doses"]):
            for j, y in enumerate(x):
                d["doses"][i][j] = DosBySpinEnergy.from_dict(y)

        return cls(**d)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import os
from argparse import Namespace
from copy import deepcopy
from pathlib import Path

import pytest
from pymatgen import Structure, Element
from vise.analyzer.atom_grouping_type import AtomGroupingType
from vise.cli.main_functions import get_poscar_from_mp, VaspSet, plot_band, \
    plot_dos, band_edge_properties, make_atom_poscars
from vise.defaults import defaults
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.task import Task
from vise.input_set.xc import Xc


def test_get_poscar_from_mp(tmpdir):
    args = Namespace(mpid="mp-110",
                     poscar="POSCAR",
                     prior_info=Path("prior_info.yaml"))
    tmpdir.chdir()
    get_poscar_from_mp(args)
    expected = """Mg1
1.0
-1.789645  1.789645  1.789645
 1.789645 -1.789645  1.789645
 1.789645  1.789645 -1.789645
1
direct
0.000000 0.000000 0.000000 Mg
"""
    assert Structure.from_file("POSCAR") == Structure.from_str(expected,
                                                               fmt="POSCAR")
    assert Path("prior_info.yaml").read_text() == """band_gap: 0.0
data_source: mp-110
icsd_ids:
- 180455
- 642652
total_magnetization: 0.0001585
"""
    # Need to remove file to avoid the side effect for other unittests.
    os.remove("prior_info.yaml")


def test_make_atom_poscars(mocker):
    args = Namespace(dirname=Path("a"), elements=[Element.H, Element.He])
    mock = mocker.patch("vise.cli.main_functions.make_atom_poscar_dirs")
    make_atom_poscars(args)
    mock.assert_called_once_with(Path("a"), [Element.H, Element.He])


default_option_args = {"poscar": "POSCAR",
                       "task": Task.structure_opt,
                       "xc": Xc.pbe,
                       "kpt_density": 1.0,
                       "overridden_potcar": ["Mn_pv"]}

default_args = deepcopy(default_option_args)
default_args.update({"user_incar_settings": None,
                     "prev_dir": None,
                     "options": None,
                     "file_transfer_type": None,
                     "uniform_kpt_mode": False,
                     "vasprun": Path("vasprun.xml"),
                     "outcar": Path("outcar.xml"),
                     })


test_data = [
    ({}, {}, {}, {}),

    ({"user_incar_settings": ["ISPIN", "2"]},
     {"ISPIN": 2},
     {},
     {}),

    ({"options": ["only_even_num_kpts", "True"]},
     {},
     {"only_even_num_kpts": True},
     {}),

    ({"uniform_kpt_mode": True},
     {},
     {"kpt_mode": KpointsMode.uniform},
     {}),

    ({"prev_dir": Path("a"), "file_transfer_type": ["file", "c"]},
     {},
     {},
     {"x": "y"})
]


@pytest.mark.parametrize("modified_settings,"
                         "overridden_incar_settings,"
                         "overridden_options_args,"
                         "prior_info", test_data)
def test_user_incar_settings(mocker,
                             modified_settings,
                             overridden_incar_settings,
                             overridden_options_args,
                             prior_info):
    args = deepcopy(default_args)
    args.update(modified_settings)

    structure = mocker.patch("vise.cli.main_functions.Structure")
    prior_info_mock = mocker.patch(
        "vise.cli.main_functions.prior_info_from_calc_dir")
    options = mocker.patch("vise.cli.main_functions.CategorizedInputOptions")
    vif = mocker.patch("vise.cli.main_functions.VaspInputFiles")

    prior_info_mock.return_value.input_options_kwargs = prior_info

    name_space = Namespace(**args)
    VaspSet(name_space)

    option_args = deepcopy(default_option_args)
    option_args.update(overridden_options_args)
    option_args.update(prior_info)
    option_args["overridden_potcar"] = {"Mn": "Mn_pv"}
    option_args.pop("poscar")
    option_args["structure"] = structure.from_file.return_value

    options.assert_called_once_with(**option_args)

    incar_settings = defaults.user_incar_settings
    incar_settings.update(overridden_incar_settings)

    vif.assert_called_once_with(options.return_value, incar_settings)


def test_plot_band(tmpdir, test_data_files):
    tmpdir.chdir()
    args = Namespace(vasprun=test_data_files / "KO2_band_vasprun.xml",
                     kpoints_filename=str(test_data_files / "KO2_band_KPOINTS"),
                     y_range=[-10, 10],
                     filename="test.pdf")

    plot_band(args)


def test_plot_dos(tmpdir, test_data_files):
    tmpdir.chdir()  # comment out when one wants to see the figure
    args = Namespace(vasprun=test_data_files / "MnO_uniform_vasprun.xml",
                     outcar=test_data_files / "MnO_uniform_OUTCAR",
                     type=AtomGroupingType.non_equiv_sites,
                     legend=True,
                     base_energy=None,
                     crop_first_value=True,
                     x_range=[-5, 5],
                     y_max_ranges=[10, 5, 7],
                     target=["Mn", "O"],
                     filename="test.pdf")
    plot_dos(args)


def test_band_edge_info(test_data_files):
    args = Namespace(vasprun=test_data_files / "MnO_uniform_vasprun.xml",
                     outcar=test_data_files / "MnO_uniform_OUTCAR")
    band_edge_properties(args)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import re
from copy import deepcopy
from typing import Tuple, List
import numpy as np
from pymatgen import Spin, Lattice

from pymatgen.electronic_structure.plotter import BSPlotter
from pymatgen.io.vasp import Vasprun, Kpoints
from pymatgen.util.string import latexify
from vise.analyzer.plot_band import BandPlotInfo, BandInfo, XTicks, BandEdge
from vise.analyzer.plot_brillouin_zone import BZPlotInfo


def greek_to_unicode(label: str) -> str:
    d = {"GAMMA": "Γ", "SIGMA": "Σ", "DELTA": "Δ"}
    for k, v in d.items():
        label = label.replace(k, v)
    return label


def italic_to_roman(label: str) -> str:
    return re.sub(r"([A-Z])_([0-9])", r"{\\rm \1}_\2", label)


class BandPlotInfoFromVasp:
    def __init__(self,
                 vasprun: Vasprun,
                 kpoints_filename: str,
                 vasprun2: Vasprun = None,
                 energy_window: List[float] = None):
        self.vasprun = vasprun
        self.kpoints_filename = kpoints_filename
        self.vasprun2 = vasprun2
        self.energy_window = energy_window
        self.bs = self.vasprun.get_band_structure(self.kpoints_filename,
                                                  line_mode=True)

    def make_band_plot_info(self):
        bs_plotter = BSPlotter(self.bs)
        plot_data = bs_plotter.bs_plot_data(zero_to_efermi=False)
        distances = [list(d) for d in plot_data["distances"]]
        self._composition = self.vasprun.final_structure.composition

        band_info = [BandInfo(band_energies=self._remove_spin_key(plot_data),
                              band_edge=self._band_edge(self.bs, plot_data),
                              fermi_level=self.bs.efermi)]

        if self.vasprun2:
            bs2 = self.vasprun2.get_band_structure(self.kpoints_filename,
                                                   line_mode=True)
            plot_data2 = BSPlotter(bs2).bs_plot_data(zero_to_efermi=False)
            band_info.append(
                BandInfo(band_energies=self._remove_spin_key(plot_data2),
                         band_edge=self._band_edge(bs2, plot_data2),
                         fermi_level=self.bs.efermi))

        x = bs_plotter.get_ticks_old()
        x_ticks = XTicks(_sanitize_labels(x["label"]), x["distance"])

        return BandPlotInfo(band_info_set=band_info,
                            distances_by_branch=distances,
                            x_ticks=x_ticks,
                            title=self._title)

    def make_bz_plot_info(self):
        rec_lat = self.vasprun.final_structure.lattice.reciprocal_lattice
        faces = [[[float(k) for k in j] for j in i] for i in rec_lat.get_wigner_seitz_cell()]
        labels = {}

        concat = False
        band_paths = []
        init_point = None
        for kpoint in self.bs.kpoints:
            if kpoint.label:
                c_coords = list(kpoint.cart_coords)
                f_coords = list(kpoint.frac_coords)
                label = greek_to_unicode(kpoint.label)
                labels[label] = {"cart": c_coords, "frac": f_coords}
                if concat is False and init_point:
                    band_paths.append([init_point, c_coords])
                init_point = c_coords
                concat = True
            else:
                concat = False

        return BZPlotInfo(faces, labels, band_paths, rec_lat.matrix.tolist())

    def _remove_spin_key(self, plot_data) -> List[List[List[List[float]]]]:
        """
        Pymatgen at 2020.11.11
         energy: A dict storing bands for spin up and spin down data
            {Spin:[np.array(nb_bands,kpoints),...]} as a list of discontinuous kpath
            of energies. The energy of multiple continuous branches are stored together.

        -> [branch][spin][band][k-point]
        """
        num_spin = len(plot_data["energy"])
        num_branch = len(plot_data["energy"]["1"])

        result = [[[] for _ in range(num_spin)] for __ in range(num_branch)]
        for spin_idx, (_, branch_energies) in enumerate(
                sorted(plot_data["energy"].items(),
                       key=lambda item: item[0],
                       reverse=True)):
            for branch_idx, branch_energy in enumerate(branch_energies):
                if self.energy_window:
                    removed_idxs = []
                    for i in range(len(branch_energy)):
                        _max = np.max(branch_energy[i, :])
                        _min = np.min(branch_energy[i, :])
                        if not self.in_energy(_max, _min):
                            removed_idxs.append(i)
                    x = np.delete(branch_energy, removed_idxs, axis=0).tolist()
                else:
                    x = branch_energy.tolist()

                result[branch_idx][spin_idx] = deepcopy(x)

        return result

    def in_energy(self, _max, _min):
        return _max >= self.energy_window[0] and _min <= self.energy_window[1]

    def _band_edge(self, bs, plot_data):
        if bs.is_metal():
            return None
        else:
            return BandEdge(
                vbm=plot_data["vbm"][0][1],
                cbm=plot_data["cbm"][0][1],
                vbm_distances=[i[0] for i in plot_data["vbm"]],
                cbm_distances=[i[0] for i in plot_data["cbm"]])

    @property
    def _title(self):
        return latexify(self._composition.reduced_formula)


def _sanitize_label(label):
    return italic_to_roman(greek_to_unicode(label))


def _sanitize_labels(labels):
    return [_sanitize_label(label) for label in labels]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import argparse
import sys
import warnings
from pathlib import Path

from pymatgen import Element
from pymatgen.io.vasp.inputs import UnknownPotcarWarning
from vise import __version__
from vise.analyzer.atom_grouping_type import AtomGroupingType
from vise.cli.main_functions import get_poscar_from_mp, VaspSet, plot_band, \
    plot_dos, band_edge_properties, make_atom_poscars, plot_absorption, \
    calc_effective_mass
from vise.defaults import defaults
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.logger import get_logger
from vise.util.str_related_tools import str2bool

logger = get_logger(__name__)


warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args(args):
    vise_yaml_files = '\n'.join(["* " + str(f) for f in defaults.yaml_files])
    parser = argparse.ArgumentParser(
        description=f"""
Vise is a package that helps researchers to do first-principles calculations
with the VASP code.

Author: Yu Kumagai
Version: {__version__}
    """,
        epilog=f"The parsed vise.yaml files are:\n{vise_yaml_files}",
        formatter_class=argparse.RawTextHelpFormatter)

    subparsers = parser.add_subparsers()

    # -- parent parser: vasprun
    vasprun_parser = argparse.ArgumentParser(description="", add_help=False)
    vasprun_parser.add_argument(
        "-v", "--vasprun", type=Path, default=defaults.vasprun,
        help="vasprun.xml file name.")
    # -- parent parser: outcar
    outcar_parser = argparse.ArgumentParser(description="", add_help=False)
    outcar_parser.add_argument(
        "-o", "--outcar", type=Path, default=defaults.outcar,
        help="OUTCAR file name.")

    # -- get_poscar -----------------------------------------------------------
    parser_get_poscar = subparsers.add_parser(
        name="get_poscar",
        description="Tools for generating a POSCAR file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['gp'])

    parser_get_poscar.add_argument(
        "-p", "--poscar", type=str, default="POSCAR",
        help="POSCAR-type file name.", metavar="FILE")
    parser_get_poscar.add_argument(
        "-pi", "--prior_info", type=Path, default=Path("prior_info.yaml"),
        help="prior_info.yaml file name.", metavar="FILE")
    parser_get_poscar.add_argument(
        "-m", "--mpid", type=str, required=True,
        help="MP entry id with prefix, e.g., mp-1234.")

    parser_get_poscar.set_defaults(func=get_poscar_from_mp)

    # -- make_atom_poscars -----------------------------------------------------
    parser_make_atom_poscars = subparsers.add_parser(
        name="make_atom_poscars",
        description="Tools for generating POSCAR files for atom calculations.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['map'])

    parser_make_atom_poscars.add_argument(
        "-d", "--dirname", type=Path, default=Path.cwd(),
        help="Directory name where atom calculation directories are created.")
    parser_make_atom_poscars.add_argument(
        "-e", "--elements", type=Element, default=None, nargs="+",
        help="Element names. When not set, all atom directories are created")

    parser_make_atom_poscars.set_defaults(func=make_atom_poscars)

    # -- vasp_set ---------------------------------------------------------
    parser_vasp_set = subparsers.add_parser(
        name="vasp_set",
        description="Tools for constructing vasp input set",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['vs'])

    parser_vasp_set.add_argument(
        "-p", "--poscar", default="POSCAR", type=Path,
        help="POSCAR-type input structure file name.")
    parser_vasp_set.add_argument(
        "-t", dest="task", default=defaults.task, type=Task,
        choices=Task.name_list(),
        help=f"Task name.")
    parser_vasp_set.add_argument(
        "-x", dest="xc", default=defaults.xc, type=Xc, choices=Xc.name_list(),
        help=f"Exchange-correlation (XC) interaction.")
    parser_vasp_set.add_argument(
        # Default must be None to use the insulator_kpoint_density.
        # defaults.kpoints_density is used at structure_kpoints_generator.py
        "-k", "--kpt_density", type=float,
        help="K-point density in Angstrom along each direction. "
             f"Defaults of metal and insulators are {defaults.kpoint_density}, "
             f"and {defaults.insulator_kpoint_density}.")
    parser_vasp_set.add_argument(
        "--potcar", dest="overridden_potcar",
        default=defaults.overridden_potcar, type=str, nargs="+",
        help="User specifying POTCAR set. E.g., Mg_pv O_h")
    parser_vasp_set.add_argument(
        "-uis", "--user_incar_settings", type=str, nargs="+",
        help="""Overwritten incar settings described with pairs of INCAR tag 
        names and values. These can also be set by vise.yaml, but they are 
        overwritten by this command line options.""")
    parser_vasp_set.add_argument(
        "-d", "--prev_dir", type=Path,
        help="Parse the previous calculations for better input settings.")
    parser_vasp_set.add_argument(
        "--options", type=str, nargs="+",
        help="Manual options used at CategorizedInputOptions class.")
    parser_vasp_set.add_argument(
        "--uniform_kpt_mode", action="store_true",
        help="""Kpoints with uniform k-point sampling. The k-point sampling mesh 
        and centering are determined based on the given lattice. Note that only 
        when the angles are 90 degrees, the centering is shifted along the 
        perpendicular direction. This mode is useful when calculating the 
        supercells. """)
    parser_vasp_set.add_argument(
        "--file_transfer_type", type=str, nargs="+",
        help="File transfer which is used with the prev_dir argument and can "
             "be written with pairs of file names and transfer types of "
             "m (move), c (copy) or l (link). An example is: "
             "--file_transfer_type POSCAR c WAVECAR l ")

    parser_vasp_set.set_defaults(func=VaspSet)

    # -- plot_band -------------------------------------------------------------
    parser_plot_band = subparsers.add_parser(
        name="plot_band",
        description="Tools for plotting band structures",
        parents=[vasprun_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pb'])

    parser_plot_band.add_argument(
        "-k", "--kpoints", dest="kpoints_filename", default="KPOINTS", type=str,
        help="KPOINTS file name.")
    parser_plot_band.add_argument(
        "-y", "--y_range", nargs="+", default=[-10, 10], type=float,
        help="Energy range, requiring two values.")
    parser_plot_band.add_argument(
        "-f", "--filename", type=str, default="band.pdf", help="Pdf file name.")

    parser_plot_band.set_defaults(func=plot_band)

    # -- plot_dos -----------------------------------------------------------
    parser_plot_dos = subparsers.add_parser(
        name="plot_dos",
        description="Tools for plotting density of states",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pd'])

    parser_plot_dos.add_argument(
        "-t", "--type",
        type=AtomGroupingType,
        default=AtomGroupingType.non_equiv_sites,
        choices=AtomGroupingType.name_list(),
        help="How to group atoms for pdos.")
    parser_plot_dos.add_argument(
        "-l", "--legend", type=str2bool, default=True,
        help="Whether to show the figure legend.")
    parser_plot_dos.add_argument(
        "-c", "--crop_first_value", type=str2bool, default=True,
        help="Whether to crop the first value in DOS.")
    parser_plot_dos.add_argument(
        "-x", "--x_range", nargs="+", type=float,
        help="Set minimum and maximum energies.")
    parser_plot_dos.add_argument(
        "-y", "--y_max_ranges", nargs="+", type=float,
        help="Set max energies for each plot. AssertionError is raised when"
             "the number of ranges does not fit to the number of figures.")
    parser_plot_dos.add_argument(
        "--target", type=str, nargs="+",
        help="""Show specific PDOS. The input depends on AtomGroupingType.\n 
        AtomGroupingType.atoms:   1,2 3 \n
        AtomGroupingType.elements: Mg O""")
    parser_plot_dos.add_argument(
        "-f", "--filename", type=str, default="dos.pdf", help="Pdf file name.")
    parser_plot_dos.add_argument(
        "-b", "--base_energy", type=float,
        help="Set when showing the figure in the absolute energies scale.")
    parser_plot_dos.set_defaults(func=plot_dos)

    # -- plot_absorption -------------------------------------------------------
    parser_plot_absorption = subparsers.add_parser(
        name="plot_absorption",
        description="Tools for plotting absorption coefficient",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pa'])

    parser_plot_absorption.add_argument(
        "-f", "--filename", type=str, default="absorption.pdf",
        help="Pdf file name.")
    parser_plot_absorption.set_defaults(func=plot_absorption)

    # -- effective_mass -------------------------------------------------------
    parser_effective_mass = subparsers.add_parser(
        name="effective_mass",
        description="Tools for calculating effective masses",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['em'])

    parser_effective_mass.add_argument(
        "-t", "--temperature", type=float, default=300, help="Temperature.")
    parser_effective_mass.add_argument(
        "-c", "--concentrations", type=lambda x: 10 ** float(x),
        nargs="+", default=18,
        help="Exponential parts of base-10 for concentrations.")

    parser_effective_mass.set_defaults(func=calc_effective_mass)

    # -- band_edge -------------------------------------------------------------
    parser_band_edge = subparsers.add_parser(
        name="band_edge",
        description="Calculate the band edge positions",
        parents=[vasprun_parser, outcar_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['be'])

    parser_band_edge.set_defaults(func=band_edge_properties)

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
        # https://github.com/kislyuk/argcomplete#activating-global-completion
        # This supports bash/zsh autocompletion. To enable this, pip install
        # argcomplete, activate global completion, or add
        #      eval "$(register-python-argcomplete vise)"
        # into your .bash_profile or .bashrc
    except ImportError:
        pass

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.dielectric_function import DieleFuncData
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties

import numpy as np


def make_diele_func(vasprun: Vasprun, outcar: Outcar):
    energies, real, imag = vasprun.dielectric_data["density"]
    real, imag = np.array(real), np.array(imag)
    band_gap = VaspBandEdgeProperties(vasprun, outcar).band_gap
    return DieleFuncData(energies, real.tolist(), imag.tolist(), band_gap)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
from plotly.graph_objs import Figure


def show_png(fig: Figure):
    fig.write_image("fig.png")
    img = mpimg.imread('fig.png')
    imageplot = plt.imshow(img)
    plt.show()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
import dash
from pydefect.analyzer.band_edge_states import BandEdgeEigenvalues
from pydefect.analyzer.defect_energy import DefectEnergy
from pydefect.analyzer.defect_energy_plotter import DefectEnergyPlotlyPlotter
from pydefect.analyzer.eigenvalue_plotter import EigenvaluePlotlyPlotter
from pydefect.corrections.efnv_correction.efnv_correction import PotentialSite, \
    ExtendedFnvCorrection
from pydefect.corrections.efnv_correction.site_potential_plotter import \
    SitePotentialPlotlyPlotter
from pymatgen import loadfn, Structure, Lattice
from pymatgen.util.string import latexify
from vise.analyzer.dash_components.structure_component import StructureComponent
from vise.analyzer.dos_data import DosData
import dash_html_components as html

import crystal_toolkit.components as ctc

app = dash.Dash()
ctc.register_app(app)

structure = Structure(lattice=Lattice.cubic(10), species=["H"], coords=[[0.0]*3])

component = StructureComponent(structure)
my_layout = html.Div(
    [
        component.layout(),
    ]
)

# app.layout = ctc.crystal_toolkit_layout(my_layout)
ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)


if __name__ == "__main__":
    app.run_server(debug=True, port=8051)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass

from monty.json import MSONable
from monty.serialization import loadfn
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class TestTest(MSONable, ToJsonFileMixIn):
    a: str = "aaa"


def test_to_json_file_mix_in(tmpdir):
    actual = TestTest()
    tmpdir.chdir()
    actual.to_json_file()
    actual = loadfn("test_test.json")
    assert actual == TestTest()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy
from dataclasses import dataclass
from typing import List, Dict, Optional

import numpy as np
from monty.json import MSONable
from pymatgen.electronic_structure.core import Spin

from vise.defaults import defaults
from vise.util.logger import get_logger

logger = get_logger(__name__)


@dataclass
class BandEdge(MSONable):
    energy: float
    spin: Spin = None
    band_index: int = None
    kpoint_index: int = None
    kpoint_coords: List[float] = None

    def is_direct(self, other: "BandEdge"):
        return (self.spin == other.spin and
                self.kpoint_coords == other.kpoint_coords)

    def __repr__(self):
        kpt_coords = f"{self.kpoint_coords[0]:5.3f} " \
                     f"{self.kpoint_coords[1]:5.3f} " \
                     f"{self.kpoint_coords[2]:5.3f}"
        return f"energy position: {self.energy}, spin: {self.spin.name:>4}, " \
               f"band index {self.band_index}, " \
               f"k-point index {self.kpoint_index}, k-point coords {kpt_coords}"

    def as_dict(self):
        d = {"@module":       self.__class__.__module__,
             "@class":        self.__class__.__name__,
             "energy":        self.energy,
             "spin":          int(self.spin),
             "band_index":    self.band_index,
             "kpoint_index":  self.kpoint_index,
             "kpoint_coords": self.kpoint_coords}

        return d

    @classmethod
    def from_dict(cls, d):
        kwargs = deepcopy(d)
        kwargs["spin"] = Spin(d["spin"])
        kwargs.pop("@module", None)
        kwargs.pop("@class", None)
        return cls(**kwargs)


class BandEdgeProperties:
    def __init__(self,
                 eigenvalues: Dict[Spin, np.ndarray],
                 nelect: float,
                 magnetization: float,
                 kpoints: List[List[float]],
                 integer_criterion: float = defaults.integer_criterion):

        assert 0 < integer_criterion < 0.5

        # [Spin][k-point_idx][band_idx] = eigenvalue
        self._eigenvalues = eigenvalues
        self._nelect = nelect
        #  In Bohr magneton.
        self._magnetization = magnetization
        self._kpoints = kpoints
        self._integer_criterion = integer_criterion

        self._calculate_vbm_cbm()

        if self._is_metal():
            self.vbm_info = None
            self.cbm_info = None

    def _calculate_vbm_cbm(self):
        self.vbm_info = BandEdge(float("-inf"))
        self.cbm_info = BandEdge(float("inf"))
        for spin, eigenvalues in self._eigenvalues.items():
            # ho = highest occupied
            ho_eigenvalue = np.amax(eigenvalues[:, self._ho_band_index(spin)])
            if ho_eigenvalue > self.vbm_info.energy:
                self.vbm_info = self.band_edge(
                    eigenvalues, self._ho_band_index(spin), ho_eigenvalue, spin)

            # lu = lowest unoccupied
            lu_band_index = self._ho_band_index(spin) + 1
            lu_eigenvalue = np.amin(eigenvalues[:, lu_band_index])
            if lu_eigenvalue < self.cbm_info.energy:
                self.cbm_info = self.band_edge(
                    eigenvalues, lu_band_index, lu_eigenvalue, spin)

    def band_edge(self, eigenvalues, band_index, eigenvalue, spin):
        k_index = int(np.where(eigenvalues[:, band_index] == eigenvalue)[0][0])
        return BandEdge(eigenvalue, spin, band_index, k_index,
                        self._kpoints[k_index])

    def _ho_band_index(self, spin):
        if spin == Spin.up:
            num_occupied_band = (self._nelect + self._magnetization) / 2
        else:
            num_occupied_band = (self._nelect - self._magnetization) / 2

        return round(num_occupied_band) - 1

    def _is_metal(self):
        nelect_frac = abs(round(self._nelect) - self._nelect)
        is_nelect_frac = nelect_frac > self._integer_criterion

        mag_frac = abs(round(self._magnetization) - self._magnetization)
        is_mag_frac = mag_frac > self._integer_criterion

        is_vbm_higher_than_cbm = self.vbm_info.energy > self.cbm_info.energy

        return is_nelect_frac or is_mag_frac or is_vbm_higher_than_cbm

    @property
    def is_metal(self):
        return self.vbm_info is None

    @property
    def is_direct(self):
        return self.vbm_info.is_direct(self.cbm_info) if self.vbm_info else None

    @property
    def band_gap(self):
        return self.cbm_info.energy - self.vbm_info.energy if self.vbm_info else None

    @property
    def vbm_cbm(self):
        return [self.vbm_info.energy, self.cbm_info.energy] if self.vbm_info else None

    def __repr__(self):
        if self.vbm_info is None:
            return "Metal"

        lines = [f"Band gap {self.band_gap:5.3f} eV",
                 f"VBM {self.vbm_info}",
                 f"CBM {self.cbm_info}"]

        return "\n".join(lines)


def is_band_gap(band_gap: Optional[float],
                vbm_cbm: Optional[List[float]]) -> bool:
    if not band_gap:
        band_gap = vbm_cbm[1] - vbm_cbm[0] if vbm_cbm else None
    if band_gap:
        logger.info(f"Band gap: {round(band_gap, 3)} eV.")
        return band_gap > defaults.band_gap_criterion
    return False



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy
from dataclasses import dataclass
from itertools import cycle
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
from monty.json import MSONable
from num2words import num2words

from vise.error import ViseError
from vise.util.matplotlib import float_to_int_formatter
from vise.util.mix_in import ToJsonFileMixIn


@dataclass(frozen=True)
class XTicks(MSONable):
    labels: List[str]
    distances: List[float]


@dataclass
class BandEdge(MSONable):
    vbm: float
    cbm: float
    vbm_distances: List[float]
    cbm_distances: List[float]


class BandInfo(MSONable):
    def __init__(self,
                 # [branch][spin][band][k-point]
                 band_energies: List[List[List[List[float]]]],
                 band_edge: Optional[BandEdge] = None,
                 fermi_level: Optional[float] = None):
        self.band_energies = deepcopy(band_energies)
        self.band_edge = deepcopy(band_edge)
        self.fermi_level = fermi_level

        if self.band_edge is None and self.fermi_level is None:
            raise ViseBandInfoError

    def slide_energies(self, base_energy):
        self._slide_band_energies(base_energy)
        self._slide_band_edge(base_energy)
        self._slide_fermi_level(base_energy)

    def _slide_band_energies(self, base_energy):
        new_array = []
        for band_energies_each_branch in self.band_energies:
            inner_array = []
            for band_energies_each_band in band_energies_each_branch:
                a = np.array(band_energies_each_band)
                inner_array.append((a - base_energy).tolist())
            new_array.append(inner_array)
        self.band_energies = new_array

    def _slide_fermi_level(self, base_energy):
        if self.fermi_level:
            self.fermi_level -= base_energy

    def _slide_band_edge(self, base_energy):
        if self.band_edge:
            self.band_edge.vbm -= base_energy
            self.band_edge.cbm -= base_energy

    @property
    def is_magnetic(self):
        return len(self.band_energies[0]) == 2


class ViseBandInfoError(ViseError):
    pass


class BandMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 linewidth: List[float] = None,
                 circle_size: int = 70,
                 circle_colors: Optional[List[str]] = None,
                 band_edge_line_width: float = 0.75,
                 band_edge_line_color: str = "black",
                 band_edge_line_style: str = "-.",
                 title_font_size: int = 15,
                 label_font_size: int = 15,
                 show_legend: bool = True,
                 legend_location: str = "lower right"
                 ):
        self.colors = colors or ['#E15759', '#4E79A7', '#F28E2B', '#76B7B2']
        self.linewidth = cycle(linewidth) if linewidth else cycle([1.0])

        self.circle_size = circle_size
        self.hline = {"linewidth": band_edge_line_width,
                      "color": band_edge_line_color,
                      "linestyle": band_edge_line_style}

        self.title_font_size = title_font_size
        self.label_font_size = label_font_size

        self.circle_colors = circle_colors or ["pink", "green"]
        self.circle_size = circle_size

        self.show_legend = show_legend
        self.legend = {"loc": legend_location}

    def band_structure(self, index):
        result = {"color": self.colors[index],
                  "linewidth": next(self.linewidth)}
        if index > 0:
            result["label"] = num2words(index + 1, ordinal=True)
        return result

    def circle(self, index):
        return {"color": self.colors[index],
                "marker": "o",
                "s": self.circle_size}


@dataclass
class BandPlotInfo(MSONable, ToJsonFileMixIn):
    band_info_set: List[BandInfo]
    distances_by_branch: List[List[float]]
    x_ticks: XTicks
    title: str = None

    def __post_init__(self):
        assert self.distances_by_branch[0][0] == self.x_ticks.distances[0]
        assert self.distances_by_branch[-1][-1] == self.x_ticks.distances[-1]

    def __add__(self, other: "BandPlotInfo"):
        assert self.distances_by_branch == other.distances_by_branch
        new_band_info_set = self.band_info_set + other.band_info_set
        return BandPlotInfo(new_band_info_set, self.distances_by_branch,
                            self.x_ticks, self.title)


class BandPlotter:

    def __init__(self,
                 band_plot_info: BandPlotInfo,
                 energy_range: List[float],
                 base_energy: float = None,
                 mpl_defaults: Optional[BandMplSettings] = BandMplSettings()
                 ):

        self.band_info_set = band_plot_info.band_info_set
        self.distances_by_branch = band_plot_info.distances_by_branch
        self.x_ticks = band_plot_info.x_ticks

        self.energy_range = energy_range
        self.title = band_plot_info.title
        self.mpl_defaults = mpl_defaults
        self.plt = plt

        self._slide_energies(base_energy)

    def _slide_energies(self, base_energy):
        if base_energy is None:
            if self.band_info_set[0].band_edge is not None:
                base_energy = self.band_info_set[0].band_edge.vbm
            elif self.band_info_set[0].fermi_level:
                base_energy = self.band_info_set[0].fermi_level

        for band_info in self.band_info_set:
            band_info.slide_energies(base_energy)

    def construct_plot(self):
        self._add_band_set()
        self._set_figure_legend()
        self._set_x_range()
        self._set_y_range()
        self._set_labels()
        self._set_x_ticks()
        self._set_title()
        self._set_formatter()
        self.plt.tight_layout()

    def _add_band_set(self):
        for index, band_info in enumerate(self.band_info_set):
            self._add_band_structures(band_info, index)

            if band_info.band_edge:
                self._add_band_edge(band_info.band_edge, index)
            else:
                if band_info.fermi_level:
                    self._add_fermi_level(band_info.fermi_level)

    def _add_band_structures(self, band_info, index):
        mpl_args = self.mpl_defaults.band_structure(index)
        for distances, energies_each_branch \
                in zip(self.distances_by_branch,  band_info.band_energies):
            for spin_index, energies_by_spin in enumerate(energies_each_branch):
                if spin_index == 0:
                    mpl_args.pop("linestyle", None)
                else:
                    mpl_args["linestyle"] = ":"

                for energies_of_a_band in energies_by_spin:
                    self.plt.plot(distances, energies_of_a_band, **mpl_args)
                    mpl_args.pop("label", None)

    def _add_band_edge(self, band_edge, index):
        self.plt.axhline(y=band_edge.vbm, **self.mpl_defaults.hline)
        self.plt.axhline(y=band_edge.cbm, **self.mpl_defaults.hline)

        for dist in band_edge.vbm_distances:
            self.plt.scatter(dist, band_edge.vbm,
                             **self.mpl_defaults.circle(index))
        for dist in band_edge.cbm_distances:
            self.plt.scatter(dist, band_edge.cbm,
                             **self.mpl_defaults.circle(index))

    def _add_fermi_level(self, fermi_level):
        self.plt.axhline(y=fermi_level, **self.mpl_defaults.hline)

    def _set_figure_legend(self):
        if self.mpl_defaults.show_legend and len(self.band_info_set) > 1:
            self.plt.legend(**self.mpl_defaults.legend)

    def _set_x_range(self):
        self.plt.xlim(self.x_ticks.distances[0], self.x_ticks.distances[-1])

    def _set_y_range(self):
        self.plt.ylim(self.energy_range[0], self.energy_range[1])

    def _set_labels(self):
        self.plt.xlabel("Wave vector", size=self.mpl_defaults.label_font_size)
        self.plt.ylabel("Energy (eV)", size=self.mpl_defaults.label_font_size)

    def _set_x_ticks(self):
        axis = self.plt.gca()
        axis.set_xticks(self.x_ticks.distances)
        axis.set_xticklabels(self.x_ticks.labels)
        for distance, label in zip(self.x_ticks.distances[1:-1],
                                   self.x_ticks.labels[1:-1]):
            linestyle = "-" if "\\mid" in label else "--"
            plt.axvline(x=distance, linestyle=linestyle)

    def _set_title(self):
        self.plt.title(self.title, size=self.mpl_defaults.title_font_size)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.yaxis.set_major_formatter(float_to_int_formatter)


